//#define LOG_NDEBUG 0
#define LOG_TAG "V4L2Device2"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <log/log.h>
#ifdef HAL3_HWC_BUFFER
#include <utils/backtrace.h>
#endif
#include <cutils/properties.h>
#include "V4L2Device2.h"
#include "misc.h"
#include "Debug.h"

#define DUMP_V4L2_BUFFER            (0)
#define USE_ION_POLL                (1)
#define V4L2_PHY_ADDR               (1)

#define MIN_DEQUEUE_THRESHOLD       (1)
#define MIN_START_STREAMING_COUNT   (0)

#define IS_COMPRESSED_BUFFER(x)     (x == ANDROID_SCALER_AVAILABLE_FORMATS_BLOB)

namespace androidcamera3
{

static const nsecs_t kWaitNs = (5 * 1E9); // XChecked: Make sure our lower speed usb camera cts normal, here set to 5 sec.
static const nsecs_t kWaitNsWatchdog = (100 * 1E9); // 100 sec

V4L2Device2::V4L2Device2(int v4l2Id, V4L2Params* params, V4L2Device2Callback* callback, int maxBufferCount,
#ifdef HAL3_HWC_BUFFER
                         GrallocWrapper* pGrallocWrapper,
#endif
                         FormatConvertHelper* pConvertHelper, bool forceAllocator, bool usePhyAddr)
    : V4L2Device2Common(v4l2Id, params, callback, maxBufferCount,
#ifdef HAL3_HWC_BUFFER
                        pGrallocWrapper,
#endif
                        pConvertHelper, forceAllocator, usePhyAddr),
      mLatestPts(0),
      mDumpCount(0),
      mReportFrameCounter(-100),
      mAllocatorsInited(false),
      mEnqueueThread(NULL),
      mDequeueThread(NULL),
      mWatchdogThread(NULL)
{
    LOG_FUNCTION_NAME();
    for (int i = 0; i < V4L2_SLOT_COUNT; i++) {
        memset(&(mSlots[i]), 0x0, sizeof(V4L2_BUF_SLOT));
        mSlots[i].id = i;
    }

    for (int i = 0; i < V4L2_SLOT_COUNT; i++) {
#ifdef HAL3_HWC_BUFFER
        mAllocators[i].pAllocator = NULL;
#endif
        mAllocators[i].mEnqueued = false;
        mAllocators[i].mValid = false;
    }
}

V4L2Device2::~V4L2Device2()
{
    LOG_FUNCTION_NAME();
    stop();
    closeDev();

    if (NULL != mWatchdogThread.get()) {
        mWatchdogThread->stop();
        mWatchdogThread->requestExitAndWait();
        mWatchdogThread.clear();
        mWatchdogThread = NULL;
        ALOGV("%s[%d]: mWatchdogThread stop", __FUNCTION__, __LINE__);
    }

    releaseAllocators();
    LOG_FUNCTION_END();
}

void V4L2Device2::initAllocators(int count)
{
    LOG_FUNCTION_NAME();
    CHECK(count <= V4L2_SLOT_COUNT);

    int deviceWidth = mParams.w;
    int deviceHeight = mParams.h;

    size_t size = ROUNDUP16(deviceWidth) * ROUNDUP16(deviceHeight) * 2;// 12 / 8; // XCheck: the size. here should be w*h*3/2
#ifdef HAL3_HWC_BUFFER
    // for MIPI workaround
    size_t allocateSize = size + (size / 8);

    ALOGD("initAllocators size : %d", allocateSize);
    int flag = IonAllocator::NONCACHED
               | IonAllocator::SCPUACC
               | IonAllocator::HWIPACC;

    for(int i = 0; i < count; i++) {

        mAllocators[i].pAllocator = (AllocatorDef*) new IonAllocator(
                                        IonAllocator::eAllocBuffer,
                                        ROUNDUP16(allocateSize),
                                        IonAllocator::MediaHeapMask,
                                        flag);

        ALOGV("%s private ION buffer vaddr %p",
              __FUNCTION__,
              mAllocators[i].pAllocator->GetVirAddr());

        mAllocators[i].mValid = true;
        mAllocators[i].mWidth = ROUNDUP16(deviceWidth);
        mAllocators[i].mHeight = ROUNDUP16(deviceHeight);
        mAllocators[i].mSize = ROUNDUP16(allocateSize);
        mAllocators[i].mFormat = mHalPixelFormat;
        mAllocators[i].mVAddr = mAllocators[i].pAllocator->GetVirAddr();
        mAllocators[i].mPhyAddr = (uint32_t) mAllocators[i].pAllocator->GetPhyAddr();

    }
#else
    size_t allocateSize = size;
    ALOGV("%s[%d]: [%dx%d], size: %d, count:%i", __FUNCTION__, __LINE__, deviceWidth, deviceHeight, allocateSize, count);
    for (int i = 0; i < count; i++) {
        mAllocators[i].mValid = true;
        mAllocators[i].mWidth = ROUNDUP16(deviceWidth);
        mAllocators[i].mHeight = ROUNDUP16(deviceHeight);
        mAllocators[i].mSize = ROUNDUP16(allocateSize);
        mAllocators[i].mFormat = mHalPixelFormat;
        mAllocators[i].mVAddr = malloc(mAllocators[i].mSize);
        mAllocators[i].mPhyAddr = 0;
    }
#endif
}

void V4L2Device2::releaseAllocators()
{
    LOG_FUNCTION_NAME();
#ifdef HAL3_HWC_BUFFER
    for(int i = 0; i < V4L2_SLOT_COUNT; i++) {
        if(mAllocators[i].pAllocator != NULL) {
            CHECK(mAllocators[i].mValid);

            ALOGV("%s private ION buffer vaddr %p",
                  __FUNCTION__,
                  mAllocators[i].pAllocator->GetVirAddr());

            mAllocators[i].pAllocator->Free();
            delete mAllocators[i].pAllocator;

            mAllocators[i].pAllocator = NULL;
            mAllocators[i].mValid = false;
        }
    }
#else
    for (int i = 0; i < V4L2_SLOT_COUNT; i++) {
        if (mAllocators[i].mValid) {
            if (mAllocators[i].mVAddr) {
                free(mAllocators[i].mVAddr);
            }
            mAllocators[i].mValid = false;
        }
    }
#endif
}

int32_t V4L2Device2::findAvailableIONSlot()
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mAllocatorsLock);
    for(int i = 0; i < V4L2_SLOT_COUNT; i++) {
        AllocatorInfo* pInfo = &(mAllocators[i]);
        if(pInfo->mValid && !pInfo->mEnqueued) {
#ifdef HAL3_HWC_BUFFER
            CHECK(pInfo->pAllocator != NULL);
#endif
            pInfo->mEnqueued = true;
            return i;
        }
    }
    return -1;
}

/**
 * a. stop V4L2
 * b. return all queued requests including
 *  should be only called when streaming is stopped
 *  1. in mRequestQueue
 *  2. in mSlots
 */
void V4L2Device2::flush(bool reportError)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);

    stop();
    CHECK((!videoIn.isStreaming));

    dumpDeviceStatus();

    int32_t frameNumber = mReportFrameCounter + 1;
    nsecs_t ptsStep = 1E9 / mParams.fps;
    nsecs_t fakePtsCounter = mLatestPts + ptsStep;

    // find min frameNumber in mSlots
    int32_t num = 0x7FFFFFFF;   // max value of a int32_t
    if(frameNumber < 0) {
        for(int i = 0; i < V4L2_SLOT_COUNT; i++) {
            if(mSlots[i].r != NULL) {
                if(num >= (int)mSlots[i].r->frameNumber) {
                    num = mSlots[i].r->frameNumber;
                }
            }
        }

        frameNumber = num;
        ALOGV("%s[%d]: init frameNumber as %d", __FUNCTION__, __LINE__, frameNumber);
    }

    // report Requests in mSlots as failed request, happened when hotplug occured.
    while(getEnqueuedSlotCount() > 0) {
        int32_t slotCount = getEnqueuedSlotCount();
        ALOGV("%s[%d]: slot count %d frameNumber %d", __FUNCTION__, __LINE__, slotCount, frameNumber);

        for(int i = 0; i < V4L2_SLOT_COUNT; i++) {
            if (mSlots[i].r != NULL && (int)mSlots[i].r->frameNumber == frameNumber) {

                mSlots[i].r->pts = fakePtsCounter;
                ALOGV("%s[%d]: submitErrorRequest %d at slot index %d", __FUNCTION__, __LINE__, frameNumber, i);

                if (reportError) {
                    mSlots[i].r->mSuccess = false;
                } else {
                    mSlots[i].r->mSuccess = true;
                }

                submitV4L2Result(mSlots[i].r);
#if (USE_ION_POLL)
                clearIONSlot(mSlots[i].r);
#else
                if(mSlots[i].r->allocator != NULL) {
                    ALOGD("XXX Free allocator %p", mSlots[i].r->allocator->GetVirAddr());
                    mSlots[i].r->allocator->Free();
                    delete mSlots[i].r->allocator;
                    mSlots[i].r->allocator = NULL;
                }
#endif
                delete mSlots[i].r;
                mSlots[i].r = NULL;

                frameNumber = frameNumber + 1;
                fakePtsCounter = fakePtsCounter + ptsStep;
            }
        }
    } // end while

    // then check mRequestQueue
    while(mRequestQueue.size() > 0) {
        Request* r = new Request();
        //memset(r,0x0,sizeof(Request));
        dequeueRequest(r);
        r->pts = fakePtsCounter;
        fakePtsCounter = fakePtsCounter + ptsStep;
        if(reportError) {
            r->mSuccess = false;
        } else {
            r->mSuccess = true;
        }

        submitV4L2Result(r);

        delete r;
        r = NULL;
    }

    LOG_FUNCTION_END();
}

status_t V4L2Device2::start()
{
    LOG_FUNCTION_NAME();
    CHECK(NULL == mEnqueueThread.get());
    CHECK(NULL == mDequeueThread.get());
    CHECK(NULL == mWatchdogThread.get());

#if 0
    mWatchdogThread = new WatchdogThread(this);
    mWatchdogThread->run("V4L2Device2::Watchdog");
#endif

    mEnqueueThread = new EnqueueThread(this);
    mEnqueueThread->run("V4L2Device2::Enqueue");

    mDequeueThread = new DequeueThread(this);
    mDequeueThread->run("V4L2Device2::Dequeue");

    return NO_ERROR;
}

status_t V4L2Device2::stop()
{
    LOG_FUNCTION_NAME();
    // make sure to end EnqueueThread first before stopStreaming
    if(NULL != mEnqueueThread.get()) {
        mEnqueueThread->stop();
        mEnqueueThread->requestExitAndWait();
        mEnqueueThread.clear();
        mEnqueueThread = NULL;
        ALOGV("%s[%d]: mEnqueueThread stop", __FUNCTION__, __LINE__);
    }

    stopStreaming();
#ifdef HAL3_HWC_BUFFER
#else
    Uninit();
#endif
    if(NULL != mDequeueThread.get()) {
        mDequeueThread->stop();
        mDequeueThread->requestExitAndWait();
        mDequeueThread.clear();
        mDequeueThread = NULL;
        ALOGV("%s[%d]: mDequeueThread stop", __FUNCTION__, __LINE__);
    }

    LOG_FUNCTION_END();
    return NO_ERROR;
}

int32_t V4L2Device2::getOccupiedRequestCount()
{
    int slots = getEnqueuedSlotCount();
    int reqs = mRequestQueue.size();

    return slots + reqs;
}

status_t V4L2Device2::findMatchedRawBufferInfo(Request* r)
{
    LOG_FUNCTION_NAME();
    int deviceWidth = mParams.w;
    int deviceHeight = mParams.h;

    CHECK(NULL == r->pMarkedBuf);

    if(mForceUseAllocatorBuf) {
        return INVALID_OPERATION;
    }

    size_t size = r->sensorBuffers->size();
    for (int i = 0; i < (int)size; i++) {
        int format = (*(r->sensorBuffers))[i].format;
        int width = (*(r->sensorBuffers))[i].width;
        int height = (*(r->sensorBuffers))[i].height;
        int gralloc_usage = (*(r->sensorBuffers))[i].gralloc_usage;
        // find first raw buffer (RTK type only), also check usage (TODO)
        ALOGV("%s[%d]: [%dx%d], format: %d, usage: %d", __FUNCTION__, __LINE__, width,  height, format, gralloc_usage);
#ifdef HAL3_HWC_BUFFER
        if((RTK_RAW_BUFFER(format)) && (width == deviceWidth) && (height == deviceHeight) && ((gralloc_usage & GRALLOC_USAGE_PRIVATE_3) != 0)) {
            buffer_handle_t* pHandle = (*(r->sensorBuffers))[i].buffer;
            r->pMarkedBuf = (StreamBuffer*) & ((*(r->sensorBuffers))[i]);

            ALOGV("%s[%d]: pMarkedBuf found %p", __FUNCTION__, __LINE__, r->pMarkedBuf);
            return OK;
        }
#else
        if((width == deviceWidth) && (height == deviceHeight)) {
            buffer_handle_t* pHandle = (*(r->sensorBuffers))[i].buffer;
            r->pMarkedBuf = (StreamBuffer*) & ((*(r->sensorBuffers))[i]);
            ALOGV("%s[%d]: pMarkedBuf found %p", __FUNCTION__, __LINE__, r->pMarkedBuf);
            return OK;
        }
#endif
    }

    return INVALID_OPERATION;
}

int32_t V4L2Device2::findAvailableSlot()
{
    LOG_FUNCTION_NAME();
    for(int i = 0; i < V4L2_SLOT_COUNT; i++) {
        if(NULL == mSlots[i].r) {
            ALOGV("%s[%d]: got free slot:%i", __FUNCTION__, __LINE__, i);
            return i;
        }
    }
    ALOGW("%s[%d]: no free slot", __FUNCTION__, __LINE__);
    return -1;
}

void V4L2Device2::handleDequeueTimeout()
{
    LOG_FUNCTION_NAME();
    if(NULL != mWatchdogThread.get()) {
        mWatchdogThread->mTerminateV4L2 = true;
        mWatchdogThread->signal();
    }
}

status_t V4L2Device2::dequeueRawBuffer(int32_t* id)
{
    LOG_FUNCTION_NAME();
#ifdef HAL3_HWC_BUFFER
    struct v4l2_buffer buf;
    int ret;

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_USERPTR;
    ret = ioctl(mFd, VIDIOC_DQBUF, &buf);

    if(ret < 0) {
        closeDev();
        ALOGE("%s[%d]: Failed", __FUNCTION__, __LINE__);
        return INVALID_OPERATION;
    }

    if ((int)buf.bytesused <= 0) {
        ALOGW("[%s]: Invalid frame[%d] size:%d",
              __FUNCTION__,
              buf.index,
              (int)buf.bytesused);
    }

    uint32_t bytes = buf.bytesused;
    int index = buf.index;
    nsecs_t pts = (nsecs_t)buf.timestamp.tv_sec * 1000000000 +
                  (nsecs_t)buf.timestamp.tv_usec * 1000;

    // tracking pts.
    mLatestPts = pts;

    //// report found slot id
    *id = index;
    // update pts from V4L2 to
    Request* r = mSlots[index].r;
    r->pts = pts;
    r->bytesused = bytes;
#else
    Mutex::Autolock l(mSlotLock);
    int ret;
    videoIn.buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    videoIn.buf.memory = V4L2_MEMORY_MMAP;
    ret = ioctl(mFd, VIDIOC_DQBUF, &(videoIn.buf));
    if (ret < 0) {
        ALOGE("%s[%d]: VIDIOC_DQBUF Failed", __FUNCTION__, __LINE__);
        closeDev();
        return INVALID_OPERATION;
    } else {
        ALOGV("%s[%d]: VIDIOC_DQBUF success, index:%d, bytes:%d", __FUNCTION__, __LINE__, videoIn.buf.index, (int)(videoIn.buf).bytesused);
    }

    if ((int)(videoIn.buf).bytesused <= 0) {
        ALOGE("%s[%d]: Invalid frame[%d] size:%d", __FUNCTION__, __LINE__, (videoIn.buf).index, (int)(videoIn.buf).bytesused);
    }

    uint32_t bytes = (videoIn.buf).bytesused;
    int index = videoIn.buf.index;
    nsecs_t pts = (nsecs_t)videoIn.buf.timestamp.tv_sec * 1000000000 + (nsecs_t)videoIn.buf.timestamp.tv_usec * 1000;

    // tracking pts.
    mLatestPts = pts;

    // report found slot id
    *id = index;
    // update pts from V4L2 to request
    Request* r = mSlots[index].r;
    if(r == NULL) {
        ALOGE("Request == NULL  index %d ", index);
    } else {
        r->pts = pts;
        r->bytesused = bytes;
        ALOGV("Request != NULL  index %d ", index);
    }
#endif

#if 0
    ALOGV("%s frameNum:%d slot id:%d pts:%lld bytes:%d IONSlot %d", __FUNCTION__,
          r->frameNumber, index, r->pts, r->bytesused, r->allocatorSlotId);
#endif

#if (DUMP_V4L2_BUFFER)
    if (mDumpCount < 30) {
        char path[64] = {'\0'};
        sprintf(path, "/tmp/v4l2-2-%d-%.8x.raw", mDumpCount, r->pMarkedBuf->y);

        ALOGV("save %s", path);
        dumpData(r->pMarkedBuf->y, r->bytesused, path, &mDumpCount);
    }
#endif

    return OK;
}

status_t V4L2Device2::enqueueRawBuffer(Request* r)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mSlotLock);

    int32_t id = findAvailableSlot();
    CHECK(id >= 0);
    CHECK(mFd != -1);
#ifdef HAL3_HWC_BUFFER
    // if pMarkedBuf is NULL, then device must allocate a ION buffer by it self
    CHECK((r->pMarkedBuf != NULL || r->allocator != NULL));

    unsigned long virAddr = 0;
    int32_t buffer_size = 0;

    char* pExtraInfo = "";
    uint32_t phyAddr = 0;

    if(r->pMarkedBuf != NULL) {
        CHECK(NULL == r->allocator);
        void* pY = r->pMarkedBuf->y;
        void* pCb = r->pMarkedBuf->cb;
        void* pCr = r->pMarkedBuf->cr;

        if(pY == NULL) {
            getBufferCbCr(
                r->pMarkedBuf->buffer,
                r->pMarkedBuf->width,
                r->pMarkedBuf->height,
                &pY,
                &pCb,
                &pCr);
        }

        virAddr = (unsigned long) pY;
        phyAddr = r->pMarkedBuf->phyAddr;

        if(phyAddr == 0) {
            getBufferPhyAddr(r->pMarkedBuf->buffer, &phyAddr);
        }

        pExtraInfo = "pMarkedBuf";
        buffer_size = r->pMarkedBuf->buf_size;
    } else {
        CHECK(NULL != r->allocator);
        virAddr = (unsigned long) r->allocatorVAddr;
        phyAddr = (uint32_t) r->allocatorPhyAddr;
        buffer_size = r->allocatorBufferSize;
        pExtraInfo = "allocator";
    }

    CHECK(virAddr != 0);
    CHECK(phyAddr != 0);
    CHECK(buffer_size != 0);

#if 0
    ALOGD("%s vaddr %p phyAddr: %.8x extra:%s",
          __FUNCTION__,
          virAddr,
          phyAddr,
          pExtraInfo);
#endif

    // save request ptr
    mSlots[id].r = r;

    int ret;
    struct v4l2_buffer* pBuf = &(mSlots[id].slot);
    memset(pBuf, 0x0, sizeof(struct v4l2_buffer));

    pBuf->index = id;
    pBuf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    pBuf->memory = V4L2_MEMORY_USERPTR;
    pBuf->m.userptr = virAddr;
    pBuf->reserved2 = phyAddr;
    pBuf->reserved = 0x0;
    pBuf->length = buffer_size;

#if 0
    ALOGD("%s  reserved2 %.8x vaddr %p length %d",
          __FUNCTION__,
          pBuf->reserved2,
          pBuf->m.userptr,
          pBuf->length);
#endif
    ret = ioctl(mFd, VIDIOC_QBUF, pBuf);

    if (ret < 0) {
        ALOGE("[%s]: QBUF Failed: %s",
              __FUNCTION__,
              strerror(errno));
        return INVALID_OPERATION;
    } else {
        mDequeueThread->signal();
        return OK;
    }
#else
    // save request ptr
    mSlots[id].r = r;

    int ret;
    videoIn.buf.index = id;
    videoIn.buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    videoIn.buf.memory = V4L2_MEMORY_MMAP;
    videoIn.buf.flags = 0;
    ret = ioctl(mFd, VIDIOC_QBUF, &(videoIn.buf));
    if (ret < 0) {
        ALOGE("%s[%d]: VIDIOC_QBUF Failed: %s, id: %i", __FUNCTION__, __LINE__, strerror(errno), id);
        return INVALID_OPERATION;
    } else {
        ALOGV("%s[%d]: VIDIOC_QBUF success, id: %d", __FUNCTION__, __LINE__, id);
        mDequeueThread->signal();
        return OK;
    }
#endif
}

int32_t V4L2Device2::getEnqueuedSlotCount()
{
    LOG_FUNCTION_NAME();
    int32_t count = 0;
    for (int i = 0; i < V4L2_SLOT_COUNT; i++) {
        if(mSlots[i].r != NULL) {
            count = count + 1;
        }
    }
    ALOGV("%s[%d]: count: %d", __FUNCTION__, __LINE__, count);
    return count;
}

void V4L2Device2::submitV4L2Result(Request* r)
{
    // keep tracking frameNumber reported
    LOG_FUNCTION_NAME();
    mReportFrameCounter = r->frameNumber;

    Request* pr = new Request();
    *pr = *r;
    mCallback->returnRequest(pr);
}

void V4L2Device2::clearIONSlot(Request* r)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mAllocatorsLock);
#ifdef HAL3_HWC_BUFFER
    if(r->allocator != NULL && r->allocatorSlotId >= 0)
#else
    if(r->allocatorSlotId >= 0)
#endif
    {
        int32_t ionSlotId = r->allocatorSlotId;
        ALOGV("%s[%d]: Free ION slot %d", __FUNCTION__, __LINE__, ionSlotId);
        CHECK(mAllocators[ionSlotId].mEnqueued);
        mAllocators[ionSlotId].mEnqueued = false;
    }
}

bool V4L2Device2::enqueueRequest(const Request& r)
{
    LOG_FUNCTION_NAME();
    {
        Mutex::Autolock l(mRequestQueueLock);
        mRequestQueue.push_back(r);
    }

    mRequestCount++;
    mEnqueueThread->signal();
    return true;
}

bool V4L2Device2::dequeueRequest(Request* r)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mRequestQueueLock);

    CHECK(mRequestQueue.size() > 0);
    *r = mRequestQueue[0];
    mRequestQueue.removeItemsAt(0);

    return true;
}
// ======================== EnqueueThread ======================== //
V4L2Device2::EnqueueThread::EnqueueThread(V4L2Device2* parent)
    : mParent(parent), mRunning(true)
{
    LOG_FUNCTION_NAME();
}

V4L2Device2::EnqueueThread::~EnqueueThread()
{
    LOG_FUNCTION_NAME();
}

void V4L2Device2::EnqueueThread::signal()
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);
    mSignal.signal();
}

void V4L2Device2::EnqueueThread::stop()
{
    LOG_FUNCTION_NAME();
    mRunning = false;
    signal();
}

bool V4L2Device2::EnqueueThread::threadLoop()
{
    ALOGV("EnqueueThread::%s[%d] Beg", __FUNCTION__, __LINE__);
    // wait condition:
    // 1. no request
    // 2. no free slot
    while( (mParent->mRequestQueue.size() <= 0 || mParent->findAvailableSlot() < 0) && mRunning) {
        Mutex::Autolock l(mLock);
        mSignal.waitRelative(mLock, kWaitNs);

        ALOGV("EnqueueThread::%s[%d]: EnqueueThread wait condition ReqQueue:%d Slot:%d AccReq:%lld", __FUNCTION__, __LINE__,
              mParent->mRequestQueue.size(), mParent->findAvailableSlot(), mParent->mRequestCount);
    }

    if(!mRunning) {
        return false;
    }

    if( mParent->mRequestCount >= MIN_START_STREAMING_COUNT
        && !mParent->videoIn.isStreaming) {
        ALOGV("EnqueueThread::%s[%d]: startStreaming", __FUNCTION__, __LINE__);
        mParent->startStreaming();
    }

    Request* r = new Request();

    // take a request from queue
    mParent->dequeueRequest(r);
    status_t rst = mParent->findMatchedRawBufferInfo(r);

    if (rst != OK) {
        if (NULL == r->pMarkedBuf) {
#if (USE_ION_POLL)
            if (!mParent->mAllocatorsInited) {
                int32_t count = (mParent->mMaxBufferCount + 2) > V4L2_SLOT_COUNT ? V4L2_SLOT_COUNT : (mParent->mMaxBufferCount + 2);
                ALOGV("EnqueueThread::%s[%d]: count: %d, mParent->mMaxBufferCount: %d", __FUNCTION__, __LINE__, count, mParent->mMaxBufferCount);
                mParent->initAllocators(count);
                mParent->mAllocatorsInited = true;
            }

            /**
             * no marked buffer found, then allocate a
             * ION buffer by our self.
             */
            int32_t id = mParent->findAvailableIONSlot();
            CHECK(id >= 0);

            AllocatorInfo* pInfo = &(mParent->mAllocators[id]);

            r->allocatorSlotId      = id;
#ifdef HAL3_HWC_BUFFER
            r->allocator            = pInfo->pAllocator;
#endif
            r->allocatorWidth       = pInfo->mWidth;
            r->allocatorHeight      = pInfo->mHeight;
            r->allocatorFormat      = pInfo->mFormat;
            r->allocatorBufferSize  = pInfo->mSize;
            r->allocatorVAddr       = pInfo->mVAddr;
            r->allocatorPhyAddr     = pInfo->mPhyAddr;
            ALOGV("EnqueueThread::%s[%d]: allocator[%i]: [%dx%d], format: %d, VAddr: %p, PhyAddr:%d, BufferSize:%d", __FUNCTION__, __LINE__,
                  id, r->allocatorWidth , r->allocatorHeight, r->allocatorFormat,
                  r->allocatorVAddr, r->allocatorPhyAddr, r->allocatorBufferSize);
#else
            int deviceWidth = mParent->mParams.w;
            int deviceHeight = mParent->mParams.h;

            size_t size = ROUNDUP16(deviceWidth) * ROUNDUP16(deviceHeight) * 12 / 8;

            int flag = IonAllocator::NONCACHED
                       | IonAllocator::SCPUACC
                       | IonAllocator::HWIPACC;

            {
                Mutex::Autolock ll(mParent->mAllocatorsLock);
                r->allocator = (AllocatorDef*) new IonAllocator(
                                   IonAllocator::eAllocBuffer,
                                   size,
                                   IonAllocator::MediaHeapMask,
                                   flag);
            }
            ALOGV("EnqueueThread::%s[%d]: allocate private ION buffer [%dx%d] size %d addr:%p", __FUNCTION__, __LINE__,
                  deviceWidth, deviceHeight, size, r->allocator->GetVirAddr());

            r->allocatorWidth = ROUNDUP16(deviceWidth);
            r->allocatorHeight = ROUNDUP16(deviceHeight);
            r->allocatorFormat = mParent->mHalPixelFormat;
            r->allocatorBufferSize = size;
#endif
        }
    }
    rst = mParent->enqueueRawBuffer(r);
    CHECK(OK == rst);
    ALOGV("EnqueueThread::%s[%d] End", __FUNCTION__, __LINE__);
    return true;
}


// ======================== DequeueThread ======================== //
V4L2Device2::DequeueThread::DequeueThread(V4L2Device2* parent)
    : mParent(parent), mRunning(true)
{
    LOG_FUNCTION_NAME();
}

V4L2Device2::DequeueThread::~DequeueThread()
{
    LOG_FUNCTION_NAME();
}

void V4L2Device2::DequeueThread::signal()
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);
    mSignal.signal();
}

void V4L2Device2::DequeueThread::stop()
{
    LOG_FUNCTION_NAME();
    mRunning = false;
    signal();
}

bool V4L2Device2::DequeueThread::threadLoop()
{
    ALOGV("DequeueThread::%s[%d] Beg", __FUNCTION__, __LINE__);
    // 1. check stream started or not
    // 2. check if is ready to dequeue buffer
    while (!(mParent->getEnqueuedSlotCount() >= MIN_DEQUEUE_THRESHOLD
             && mParent->videoIn.isStreaming) && mRunning) {
        Mutex::Autolock l(mLock);
        status_t res = mSignal.waitRelative(mLock, kWaitNs);
        if (TIMED_OUT == res) {
            if(mParent->videoIn.isStreaming) {
                ALOGV("DequeueThread::%s[%d]: DequeueThread wait timeout occupied slots %d videoStreaming %d", __FUNCTION__, __LINE__,
                      mParent->getEnqueuedSlotCount(), mParent->videoIn.isStreaming);
#if 0
                mRunning = false;
                mParent->handleDequeueTimeout();
#endif
            }
        }
    }

    if (!mRunning) {
        return false;
    }

    CHECK(mParent->videoIn.isStreaming);
    CHECK(mParent->getEnqueuedSlotCount() >= MIN_DEQUEUE_THRESHOLD);

    int32_t id = -1;
    status_t res = mParent->dequeueRawBuffer(&id);
    if (OK != res || -1 == id) {
        ALOGE("DequeueThread::%s[%d]: dequeueRawBuffer error, break mDequeueThread", __FUNCTION__, __LINE__);
        return false;
    }

#ifdef HAL3_HWC_BUFFER
    ALOGV("DequeueThread::%s[%d]: DequeueThread 1 mSlot[%d], request:%d", __FUNCTION__, __LINE__, id, mParent->mSlots[id].r->frameNumber);
    CHECK(mParent->mFormatConvertHelper != NULL);
    mParent->mFormatConvertHelper->convertBuffers(mParent->mSlots[id].r);
    // update security flags
    mParent->updateStreamBufFlag(mParent->mSlots[id].r);
    mParent->mFormatConvertHelper->setRtkPrivData(mParent->mSlots[id].r, mParent->mParams.fps);
#else
    if(mParent->mSlots[id].r != NULL) {
        ALOGV("DequeueThread::%s[%d]: mSlots[%i].r->frameNumber:%d", __FUNCTION__, __LINE__, id, mParent->mSlots[id].r->frameNumber);
    } else {
        ALOGV("DequeueThread::%s[%d]: mSlots[%i].r is NULL", __FUNCTION__, __LINE__, id);
    }
    CHECK(mParent->mFormatConvertHelper != NULL);
    mParent->mFormatConvertHelper->convertBuffers(mParent->mSlots[id].r, &(mParent->videoIn));
#endif
    mParent->mSlots[id].r->mSuccess = true;

    //ALOGD("\tDequeueThread 2 r:%d",mParent->mSlots[id].r->frameNumber);
#if (USE_ION_POLL)

    mParent->clearIONSlot(mParent->mSlots[id].r);

    //ALOGD("\tDequeueThread 3 r:%d",mParent->mSlots[id].r->frameNumber);
#else

    {
        Mutex::Autolock ll(mParent->mAllocatorsLock);
        // clear private ION allocator
        if(mParent->mSlots[id].r->allocator != NULL) {
            ALOGV("DequeueThread::%s[%d]: Free allocator %p", __FUNCTION__, __LINE__, mParent->mSlots[id].r->allocator->GetVirAddr());
            mParent->mSlots[id].r->allocator->Free();
            delete mParent->mSlots[id].r->allocator;
            mParent->mSlots[id].r->allocator = NULL;
        }
    }
#endif

    mParent->submitV4L2Result(mParent->mSlots[id].r);
    delete mParent->mSlots[id].r;
    mParent->mSlots[id].r = NULL;
    ALOGV("DequeueThread::%s[%d] End", __FUNCTION__, __LINE__);
    return true;
}

// ====================== WatchdogThread ======================== //
V4L2Device2::WatchdogThread::WatchdogThread(V4L2Device2* parent)
    : mTerminateV4L2(false), mParent(parent), mRunning(true)
{
    ALOGD("WatchdogThread()");
}

V4L2Device2::WatchdogThread::~WatchdogThread()
{
    ALOGD("~WatchdogThread()");
}

void V4L2Device2::WatchdogThread::signal()
{
    Mutex::Autolock l(mLock);
    mSignal.signal();
}

void V4L2Device2::WatchdogThread::stop()
{
    mRunning = false;
    signal();
}

bool V4L2Device2::WatchdogThread::threadLoop()
{
    LOG_FUNCTION_NAME();
    while(mRunning && !mTerminateV4L2) {
        Mutex::Autolock l(mLock);
        status_t res = mSignal.waitRelative(mLock, kWaitNsWatchdog);
    }

    if(mTerminateV4L2) {
        mParent->flush(false);
    }

    return false;
}

void V4L2Device2::dumpData(void* vaddr,
                           int size,
                           char* path,
                           unsigned int* pCnt)
{
    FILE* fd = fopen(path, "w");
    if(fd != NULL) {
        fwrite(vaddr, 1, size, fd);
        int rst = fclose(fd);
        ALOGV("[%s] path %s size %d rst %d", __FUNCTION__, path, size, rst);
        (*pCnt)++;
    }
}

void V4L2Device2::dumpDeviceStatus()
{
    ALOGD("mRequestCount %lld mRequestQueue %d enqueuedSlots %d",
          mRequestCount,
          mRequestQueue.size(),
          getEnqueuedSlotCount());
}

void V4L2Device2::updateStreamBufFlag(Request* r)
{
    LOG_FUNCTION_NAME();
    Buffers::iterator buf = r->sensorBuffers->begin();
    while(buf != r->sensorBuffers->end()) {
        bool valided = false; // checkHDCP(buf);
        buf->bufferSuccess = valided;
        ++buf;
    }
}

}; // namespace androidcamera3
