//#define LOG_NDEBUG 0
#define LOG_TAG "V4L2Device2Common"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <log/log.h>
#include <ui/GraphicBufferMapper.h>
#include "V4L2Device2Common.h"
#ifdef HAL3_HWC_BUFFER
#else
#include <binder/MemoryBase.h>
#include <binder/MemoryHeapBase.h>
#include <linux/videodev2.h>

#define PIXEL_FORMAT   V4L2_PIX_FMT_YUYV;
#endif
#include <ui/Rect.h>
#include "Debug.h"

namespace androidcamera3
{

V4L2Device2Common::V4L2Device2Common(int v4l2Id, V4L2Params* params, V4L2Device2Callback* callback, int maxBufferCount,
#ifdef HAL3_HWC_BUFFER
                                     GrallocWrapper* pGrallocWrapper,
#endif
                                     FormatConvertHelper* pConvertHelper, bool forceAllocator, bool usePhyAddr)
    : mV4L2Id(v4l2Id), mFd(-1), videoIn(),
#ifdef HAL3_HWC_BUFFER
      mHalPixelFormat(HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP),
#else
      mHalPixelFormat(HAL_PIXEL_FORMAT),
#endif
      mCallback(callback),
#ifdef HAL3_HWC_BUFFER
      mGrallocWrapper(pGrallocWrapper),
#endif
      mFormatConvertHelper(pConvertHelper), mMaxBufferCount(maxBufferCount), mRequestCount(0), mForceUseAllocatorBuf(forceAllocator), mUsePhyAddr(usePhyAddr)
{
    LOG_FUNCTION_NAME();
    CHECK(params != NULL);
    memcpy(&mParams, params, sizeof(V4L2Params));
    memset(&videoIn, 0x0, sizeof(vdIn));

    memset(mDeviceStr, 0x0, STR_BUF_SIZE);
    snprintf(mDeviceStr, STR_BUF_SIZE, "/dev/video%d", v4l2Id);
    ALOGV("%s[%d]: mDeviceStr: %s", __FUNCTION__, __LINE__, mDeviceStr);
    mRequestQueue.setCapacity(32);
}

V4L2Device2Common::~V4L2Device2Common()
{
    LOG_FUNCTION_NAME();
}

int V4L2Device2Common::enumFrameFmt(int fd)
{
    int ret;
    struct v4l2_fmtdesc fmt;
    memset(&fmt, 0, sizeof(fmt));

    fmt.index = 0;
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    while ((ret = ioctl(fd, VIDIOC_ENUM_FMT, &fmt)) == 0) {
        fmt.index++;
        ALOGV("%s[%d]: {pixelformat=''%c%c%c%c'', description=''%s''}", __FUNCTION__, __LINE__, fmt.pixelformat & 0xFF, (fmt.pixelformat >> 8) & 0xFF, (fmt.pixelformat >> 16) & 0xFF, (fmt.pixelformat >> 24) & 0xFF, fmt.description);
    }
    return ret;
}

status_t V4L2Device2Common::init()
{
    LOG_FUNCTION_NAME();
    V4L2Params* params = &mParams;
    const char* device = mDeviceStr;

    int ret;
    videoIn.isInited = false;

    ALOGV("%s[%d]: %s:[%dx%d] fmt:%c%c%c%c fps %d", __FUNCTION__, __LINE__,
          device, params->w, params->h, FCC2ASCII(params->v4l2format), params->fps);

    mFd = open(device, O_RDWR);
    if (mFd == -1) {
        ALOGE("%s[%d]: ERROR opening V4L2 device: %s(%s)", __FUNCTION__, __LINE__, device, strerror(errno));
        return INVALID_OPERATION;
    }

    ALOGV("%s[%d]: mFd:%d, pixelformat:%c%c%c%c", __FUNCTION__, __LINE__, mFd, FCC2ASCII(params->v4l2format));

    ret = ioctl(mFd, VIDIOC_QUERYCAP, &(videoIn.cap));
    if (ret < 0) {
        ALOGE("Error opening device: unable to query device capabilities.");
        return INVALID_OPERATION;
    }

    // check capabilities V4L2_CAP_VIDEO_CAPTURE
    if ((videoIn.cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0) {
        ALOGE("Error opening device: video capture not supported.");
        return INVALID_OPERATION;
    }

    // check capabilities V4L2_CAP_STREAMING
    if (!(videoIn.cap.capabilities & V4L2_CAP_STREAMING)) {
        ALOGE("Capture device does not support streaming i/o");
        return INVALID_OPERATION;
    }
#ifdef HAL3_HWC_BUFFER
#else
    params->v4l2format = PIXEL_FORMAT; // XChecked: If support MJPG, here should not change format to YUYV
#endif
    // S_FMT
    videoIn.format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    videoIn.format.fmt.pix.width = params->w;
    videoIn.format.fmt.pix.height = params->h;
    videoIn.format.fmt.pix.pixelformat = params->v4l2format;
    ret = ioctl(mFd, VIDIOC_S_FMT, &(videoIn.format));
    if (ret < 0) {
        ALOGE("set format to %c%c%c%c failed: %s", FCC2ASCII(params->v4l2format), strerror(errno));
        return INVALID_OPERATION;
    }
#ifdef HAL3_HWC_BUFFER
#else
    videoIn.formatIn = params->v4l2format;
    videoIn.width = videoIn.format.fmt.pix.width;
    videoIn.height = videoIn.format.fmt.pix.height;
    // videoIn.framesizeIn = (videoIn.width * videoIn.height << 1);
    ALOGV("%s[%d]: VIDIOC_S_FMT success: [%ix%i], pixelformat:%c%c%c%c", __FUNCTION__, __LINE__, videoIn.width, videoIn.height, FCC2ASCII(params->v4l2format));
#endif

    // S_PARM
    struct v4l2_streamparm* setfps = (struct v4l2_streamparm*)calloc(1, sizeof(struct v4l2_streamparm));
    setfps->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    setfps->parm.capture.timeperframe.numerator = 1;
    videoIn.fps = params->fps;
    ALOGV("%s[%d]: Source FPS: %d", __FUNCTION__, __LINE__, videoIn.fps);
    setfps->parm.capture.timeperframe.denominator = videoIn.fps;
    ret = ioctl(mFd, VIDIOC_S_PARM, setfps);
    free(setfps);
    if (ret < 0) {
        ALOGE("%s[%d]: Setup: set framerate to %d failed: %s", __FUNCTION__, __LINE__, videoIn.fps, strerror(errno));
        return INVALID_OPERATION;
    }

    // REQ_BUF
#ifdef HAL3_HWC_BUFFER
    videoIn.rb.count = V4L2_SLOT_COUNT;
    videoIn.rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    videoIn.rb.memory = V4L2_MEMORY_USERPTR;
    ret = ioctl(mFd, VIDIOC_REQBUFS, &videoIn.rb);
    if (ret < 0) {
        ALOGE("%s[%d]: Init: VIDIOC_REQBUFS failed: %s", __FUNCTION__, __LINE__, strerror(errno));
        return INVALID_OPERATION;
    }
#else
    videoIn.rb.count = V4L2_SLOT_COUNT;
    videoIn.rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    videoIn.rb.memory = V4L2_MEMORY_MMAP;
    ret = ioctl(mFd, VIDIOC_REQBUFS, &videoIn.rb);
    if (ret < 0) {
        ALOGE("%s[%d]: Init: VIDIOC_REQBUFS failed: %s", __FUNCTION__, __LINE__, strerror(errno));
        return INVALID_OPERATION;
    }

    for (int i = 0; i < V4L2_SLOT_COUNT; i++) {
        memset (&videoIn.buf, 0, sizeof (struct v4l2_buffer));
        videoIn.buf.index = i;
        videoIn.buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        videoIn.buf.memory = V4L2_MEMORY_MMAP;
        ret = ioctl (mFd, VIDIOC_QUERYBUF, &videoIn.buf);
        if (ret < 0) {
            ALOGE("%s[%d]: Error buffer request fail index:%d, err:%s", __FUNCTION__, __LINE__, i, strerror(errno));
            return ret;
        }

        videoIn.mem[i] = mmap (0, videoIn.buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, mFd, videoIn.buf.m.offset);
        ALOGV("%s[%d]: videoIn.mem[%i]:%p, length:%i, offset:%i", __FUNCTION__, __LINE__, i, videoIn.mem[i], videoIn.buf.length, videoIn.buf.m.offset);
        if (videoIn.mem[i] == MAP_FAILED) {
            ALOGE("%s[%d]: Error Unable to map buffer (%s)", __FUNCTION__, __LINE__, strerror(errno));
            return -1;
        }
    }
#endif
    videoIn.isInited = true;
    ALOGV("%s[%d]: Init V4L2 complete", __FUNCTION__, __LINE__);
    return OK;
}


// Simple comment: this is a private function, V4L2Device2 would call in EnqueueThread
status_t V4L2Device2Common::startStreaming()
{
    LOG_FUNCTION_NAME();
    if (videoIn.isStreaming) {
        return OK;
    }

    if(videoIn.isInited == false) {
        ALOGE("%s[%d]: Start Streaming without init.", __FUNCTION__, __LINE__);
        videoIn.isStreaming = false;
        return INVALID_OPERATION;
    }

    ALOGV("%s[%d]: videoIn.buf.type:%i", __FUNCTION__, __LINE__, videoIn.buf.type);
    int ret = ioctl(mFd, VIDIOC_STREAMON, &(videoIn.buf.type));
    if (ret < 0) {
        ALOGE("%s[%d]: VIDIOC_STREAMON Failed: %s", __FUNCTION__, __LINE__, strerror(errno));
        return INVALID_OPERATION;
    }
    ALOGV("%s[%d]: Start Streaming success", __FUNCTION__, __LINE__);

    videoIn.isStreaming = true;

    return OK;
}

status_t V4L2Device2Common::stopStreaming()
{
    LOG_FUNCTION_NAME();
    if(videoIn.isInited == false || videoIn.isStreaming == false) {
        ALOGV("%s[%d]: Can not stop stream, init flag %d, stream on flag %d", __FUNCTION__, __LINE__, videoIn.isInited, videoIn.isStreaming);
        return INVALID_OPERATION;
    }
    videoIn.isStreaming = false;

    ALOGV("%s[%d]: videoIn.buf.type:%i", __FUNCTION__, __LINE__, videoIn.buf.type);
    int ret = ioctl(mFd, VIDIOC_STREAMOFF, &(videoIn.buf.type));
    if (ret < 0) {
        ALOGE("%s[%d]: StopStreaming: Unable to stop capture: %s", __FUNCTION__, __LINE__, strerror(errno));
        return INVALID_OPERATION;
    }

    return OK;
}

bool V4L2Device2Common::isStreaming()
{
    return videoIn.isStreaming;
}

#ifdef HAL3_HWC_BUFFER
#else
void V4L2Device2Common::Uninit()
{
    LOG_FUNCTION_NAME();
    for (int i = 0; i < V4L2_SLOT_COUNT; i++) {
        if (munmap(videoIn.mem[i], videoIn.buf.length) < 0) {
            ALOGE("%s[%d]: [%i] Unmap failed", __FUNCTION__, __LINE__, i);
        }
    }
    videoIn.isInited = false;
}
#endif

void V4L2Device2Common::getBufferCbCr(buffer_handle_t* buffer, int w, int h, void** ppVPtr, void** ppCb, void** ppCr)
{
    LOG_FUNCTION_NAME();
#ifdef HAL3_HWC_BUFFER
    int usage = mGrallocWrapper->GetUsage(*buffer);
#else
    int usage = CAMERA_GRALLOC_USAGE; // XChecked: usage
#endif
    const Rect rect(w, h);
    android_ycbcr ycbcr = android_ycbcr();
    memset(&ycbcr, 0x0, sizeof(android_ycbcr));
    status_t rst = GraphicBufferMapper::get().lockYCbCr(*buffer, usage, rect, &ycbcr);
    *ppVPtr = ycbcr.y;
    *ppCb = ycbcr.cb;
    *ppCr = ycbcr.cr;

    GraphicBufferMapper::get().unlock(*buffer);
}

void V4L2Device2Common::getBufferPhyAddr(buffer_handle_t* buffer, uint32_t* phyAddr)
{
    LOG_FUNCTION_NAME();
    uint32_t phy_info[2];

    phy_info[0] = 0;
    phy_info[1] = 0;

    CHECK(buffer != NULL);
#ifdef HAL3_HWC_BUFFER
    int rst = GrallocWrapper_GetPhyAddrs(mGrallocWrapper, *buffer, (uint32_t*)phy_info);
    ALOGD("getBufferPhyAddr rst %d", rst);
#endif
    *phyAddr = phy_info[0];
}

void V4L2Device2Common::closeDev()
{
    LOG_FUNCTION_NAME();
    // dequeue all buffers
    if(mFd >= 0) {
        struct v4l2_buffer buf;
        while(ioctl(mFd, VIDIOC_DQBUF, &buf) == 0);
        ALOGV("%s[%d]: mFd:%d", __FUNCTION__, __LINE__, mFd);
        close(mFd);
        mFd = -1;
    }
}
};
